# The functions here are modified from:
# https://github.com/fgnt/pb_bss/blob/master/pb_bss/distribution/complex_angular_central_gaussian.py

from dataclasses import dataclass

import cupy as cp

from cacgmm.utils import normalize_observation, is_broadcast_compatible


@dataclass
class ComplexAngularCentralGaussian:
    """
    Note:
        Instead of the covariance the eigenvectors and eigenvalues are saved.
        These saves some computations, because to have a more stable covariance,
        the eigenvalues are floored.
    """

    covariance_eigenvectors: cp.array = None  # (..., D, D)
    covariance_eigenvalues: cp.array = None  # (..., D)

    @classmethod
    def from_covariance(
        cls,
        covariance,
        eigenvalue_floor=0.0,
        covariance_norm="eigenvalue",
    ):
        if covariance_norm == "trace":
            cov_trace = cp.einsum("...dd", covariance)[..., None, None]
            covariance /= cp.maximum(cov_trace, cp.finfo(cov_trace.dtype).tiny)
        else:
            assert covariance_norm in ["eigenvalue", False]

        try:
            eigenvals, eigenvecs = cp.linalg.eigh(covariance)
        except cp.linalg.LinAlgError:
            # ToDo: figure out when this happen and why eig may work.
            # It is likely that eig is more stable than eigh.
            try:
                eigenvals, eigenvecs = cp.linalg.eig(covariance)
            except cp.linalg.LinAlgError:
                if eigenvalue_floor == 0:
                    raise RuntimeError(
                        "When you set the eigenvalue_floor to zero it can "
                        "happen that the eigenvalues get zero and the "
                        "reciprocal eigenvalue that is used in "
                        f"{cls.__name__}._log_pdf gets infinity."
                    )
                else:
                    raise
        eigenvals = eigenvals.real
        if covariance_norm == "eigenvalue":
            # The scale of the eigenvals does not matter.
            eigenvals = eigenvals / cp.maximum(
                cp.amax(eigenvals, axis=-1, keepdims=True),
                cp.finfo(eigenvals.dtype).tiny,
            )
            eigenvals = cp.maximum(
                eigenvals,
                eigenvalue_floor,
            )
        else:
            eigenvals = cp.maximum(
                eigenvals,
                cp.amax(eigenvals, axis=-1, keepdims=True) * eigenvalue_floor,
            )
        assert cp.isfinite(eigenvals).all(), eigenvals

        return cls(
            covariance_eigenvalues=eigenvals,
            covariance_eigenvectors=eigenvecs,
        )

    @property
    def covariance(self):
        return cp.einsum(
            "...wx,...x,...zx->...wz",
            self.covariance_eigenvectors,
            self.covariance_eigenvalues,
            self.covariance_eigenvectors.conj(),
            optimize="greedy",
        )

    @property
    def log_determinant(self):
        return cp.sum(cp.log(self.covariance_eigenvalues), axis=-1)

    def log_pdf(self, y):
        """
        Args:
            y: Shape (..., T, D)
        Returns:
        """
        y = normalize_observation(y)  # swap D and T dim
        log_pdf, _ = self._log_pdf(y)
        return log_pdf

    def _log_pdf(self, y):
        """Gets used by. e.g. the cACGMM.
        TODO: quadratic_form might be useful by itself
        Note: y shape is (..., D, T) and not (..., T, D) like in log_pdf
        Args:
            y: Normalized observations with shape (..., D, T).
        Returns: Affiliations with shape (..., K, T) and quadratic format
            with the same shape.
        """
        *independent, D, T = y.shape

        assert is_broadcast_compatible(
            [*independent, D, D], self.covariance_eigenvectors.shape
        ), (y.shape, self.covariance_eigenvectors.shape)

        quadratic_form = cp.maximum(
            cp.abs(
                cp.einsum(
                    "...dt,...de,...e,...ge,...gt->...t",
                    y.conj(),
                    self.covariance_eigenvectors,
                    1 / self.covariance_eigenvalues,
                    self.covariance_eigenvectors.conj(),
                    y,
                    optimize="optimal",
                )
            ),
            cp.finfo(y.dtype).tiny,
        )
        log_pdf = -D * cp.log(quadratic_form)
        log_pdf -= self.log_determinant[..., None]

        return log_pdf, quadratic_form



# The functions here are modified from:
# https://github.com/fgnt/pb_bss/blob/master/pb_bss/distribution/complex_angular_central_gaussian.py

import cupy as cp

from cacgmm.cacg import ComplexAngularCentralGaussian
from cacgmm.utils import force_hermitian, is_broadcast_compatible, normalize_observation


class ComplexAngularCentralGaussianTrainer:
    def fit(
        self,
        y,
        saliency=None,
        hermitize=True,
        covariance_norm="eigenvalue",
        eigenvalue_floor=1e-10,
        iterations=10,
    ):
        """
        Args:
            y: Should be normalized to unit norm. We normalize it anyway again.
               Shape (..., D, T), e.g. (1, D, T) for mixture models
            saliency: Shape (..., T), e.g. (K, T) for mixture models
            hermitize:
            eigenvalue_floor:
            iterations:
        Returns:
        """
        *independent, T, D = y.shape
        assert cp.iscomplexobj(y), y.dtype
        assert y.shape[-1] > 1
        y = normalize_observation(y)  # swap D and T dim

        if saliency is None:
            quadratic_form = cp.ones(*independent, T)
        else:
            raise NotImplementedError

        assert iterations > 0, iterations
        for _ in range(iterations):
            model = self._fit(
                y=y,
                saliency=saliency,
                quadratic_form=quadratic_form,
                hermitize=hermitize,
                covariance_norm=covariance_norm,
                eigenvalue_floor=eigenvalue_floor,
            )
            _, quadratic_form = model._log_pdf(y)

        return model

    def _fit(
        self,
        y,
        saliency,
        quadratic_form,
        hermitize=True,
        covariance_norm="eigenvalue",
        eigenvalue_floor=1e-10,
    ) -> ComplexAngularCentralGaussian:
        """Single step of the fit function. In general, needs iterations.
        Note: y shape is (..., D, T) and not (..., T, D) like in fit
        Args:
            y:  Assumed to have unit length.
                Shape (..., D, T), e.g. (1, D, T) for mixture models
            saliency: Shape (..., T), e.g. (K, T) for mixture models
            quadratic_form: (..., T), e.g. (K, T) for mixture models
            hermitize:
            eigenvalue_floor:
        """
        assert cp.iscomplexobj(y), y.dtype

        assert is_broadcast_compatible(y.shape[:-2], quadratic_form.shape[:-1]), (
            y.shape,
            quadratic_form.shape,
        )

        D = y.shape[-2]
        *independent, T = quadratic_form.shape

        if saliency is None:
            saliency = 1
            denominator = cp.array(T, dtype=cp.float64)
        else:
            assert y.ndim == saliency.ndim + 1, (y.shape, saliency.ndim)
            denominator = cp.einsum("...n->...", saliency)[..., None, None]

        # When the covariance matrix is zero, quadratic_form would also zero.
        # quadratic_form have to be positive
        quadratic_form = cp.maximum(
            quadratic_form,
            # Use 2 * tiny, because tiny is to small
            10 * cp.finfo(quadratic_form.dtype).tiny,
        )

        covariance = D * cp.einsum(
            "...dn,...Dn,...n->...dD",
            y,
            y.conj(),
            (saliency / quadratic_form),
            # optimize='greedy',
            # Without greedy the diagonal of the  covariance matrix is real
            # valued, otherwise only approximately real values
            # (i.e., values of 1e-17 for the imag part)
        )
        assert cp.isfinite(quadratic_form).all()
        covariance /= cp.maximum(
            denominator,
            cp.finfo(denominator.dtype).tiny,
        )
        assert covariance.shape == (*independent, D, D), (
            covariance.shape,
            (*independent, D, D),
        )

        assert cp.isfinite(covariance).all()

        if hermitize:
            covariance = force_hermitian(covariance)

        return ComplexAngularCentralGaussian.from_covariance(
            covariance,
            eigenvalue_floor=eigenvalue_floor,
            covariance_norm=covariance_norm,
        )



import cupy as cp


def _unit_norm(signal, *, axis=-1, eps=1e-4, eps_style="plus", ord=None):
    norm = cp.linalg.norm(signal, ord=ord, axis=axis, keepdims=True)
    if eps_style == "plus":
        norm = norm + eps
    elif eps_style == "max":
        norm = cp.maximum(norm, eps)
    elif eps_style == "where":
        norm = cp.where(norm == 0, eps, norm)
    else:
        assert False, eps_style
    return signal / norm


def force_hermitian(matrix):
    return (matrix + cp.swapaxes(matrix.conj(), -1, -2)) / 2


def estimate_mixture_weight(
    affiliation,
    saliency=None,
    weight_constant_axis=-1,
):
    affiliation = cp.asarray(affiliation)

    if (
        isinstance(weight_constant_axis, int)
        and weight_constant_axis % affiliation.ndim - affiliation.ndim == -2
    ):
        K = affiliation.shape[-2]
        return cp.full([K, 1], 1 / K)
    elif isinstance(weight_constant_axis, list):
        weight_constant_axis = tuple(weight_constant_axis)

    if saliency is None:
        weight = cp.mean(affiliation, axis=weight_constant_axis, keepdims=True)
    else:
        masked_affiliation = affiliation * saliency[..., None, :]
        weight = _unit_norm(
            cp.sum(masked_affiliation, axis=weight_constant_axis, keepdims=True),
            ord=1,
            axis=-2,
            eps=1e-10,
            eps_style="where",
        )

    return weight


def log_pdf_to_affiliation(
    weight,
    log_pdf,
    source_activity_mask=None,
    affiliation_eps=0.0,
):
    # Only check broadcast compatibility
    if source_activity_mask is None:
        _ = cp.broadcast_arrays(weight, log_pdf)
    else:
        _ = cp.broadcast_arrays(weight, log_pdf, source_activity_mask)

    # The value of affiliation max may exceed float64 range.
    # Scaling (add in log domain) does not change the final affiliation.
    affiliation = log_pdf - cp.amax(log_pdf, axis=-2, keepdims=True)

    cp.exp(affiliation, out=affiliation)

    # Weight multiplied not in log domain to avoid logarithm of zero.
    affiliation *= weight

    if source_activity_mask is not None:
        assert source_activity_mask.dtype == cp.bool, source_activity_mask.dtype  # noqa
        affiliation *= source_activity_mask

    denominator = cp.maximum(
        cp.sum(affiliation, axis=-2, keepdims=True),
        cp.finfo(affiliation.dtype).tiny,
    )
    affiliation /= denominator

    # Strictly, you need re-normalization after clipping. We skip that here.
    if affiliation_eps != 0:
        affiliation = cp.clip(
            affiliation,
            affiliation_eps,
            1 - affiliation_eps,
        )

    return affiliation


def is_broadcast_compatible(*shapes):
    if len(shapes) < 2:
        return True
    else:
        for dim in zip(*[shape[::-1] for shape in shapes]):
            if len(set(dim).union({1})) <= 2:
                pass
            else:
                return False
        return True


def normalize_observation(observation):
    """
    Attention: swap D and T dim
    """
    observation = _unit_norm(
        observation,
        axis=-1,
        eps=cp.finfo(observation.dtype).tiny,
        eps_style="where",
    )
    return cp.ascontiguousarray(cp.swapaxes(observation, -2, -1))



# The functions here are modified from:
# https://github.com/fgnt/pb_bss/blob/master/pb_bss/distribution/cacgmm.py

from operator import xor

import cupy as cp
from cacgmm.cacg_trainer import ComplexAngularCentralGaussianTrainer
from cacgmm.cacgmm import CACGMM
from cacgmm.utils import estimate_mixture_weight, normalize_observation


class CACGMMTrainer:
    def fit(
        self,
        y,
        initialization=None,
        num_classes=None,
        iterations=100,
        saliency=None,
        *,
        source_activity_mask=None,
        weight_constant_axis=(-1,),
        hermitize=True,
        covariance_norm="eigenvalue",
        affiliation_eps=1e-10,
        eigenvalue_floor=1e-10,
    ):
        """

        Args:
            y: Shape (frequency, time, channel) or (F, T, D)
            initialization:
                Affiliations between 0 and 1. Shape (F, K, T)
                or CACGMM instance
            num_classes: Scalar >0
            iterations: Scalar >0
            saliency:
                Importance weighting for each observation, shape (..., T)
                Should be pre-calculated externally, not just a string.
            source_activity_mask: Boolean mask that says for each time point
                for each source if it is active or not.
                Shape (F, K, T)
            weight_constant_axis: The axis that is used to calculate the mean
                over the affiliations. The affiliations have the
                shape (F, K, T), so the default value means averaging over
                the sample dimension. Note that averaging over an independent
                axis is supported.
            hermitize:
            covariance_norm: 'eigenvalue', 'trace' or False
            affiliation_eps:
            eigenvalue_floor: Relative flooring of the covariance eigenvalues

        Returns:

        """
        assert xor(initialization is None, num_classes is None), (
            "Incompatible input combination. "
            "Exactly one of the two inputs has to be None: "
            f"{initialization is None} xor {num_classes is None}"
        )

        assert cp.iscomplexobj(y), y.dtype
        assert y.shape[-1] > 1, y.shape
        y = normalize_observation(y)  # swap D and T dim, now y is F, D, T

        assert iterations > 0, iterations

        model = None

        *independent, D, num_observations = y.shape
        if initialization is None:
            assert num_classes is not None, num_classes
            affiliation_shape = (*independent, num_classes, num_observations)
            affiliation = cp.random.uniform(size=affiliation_shape)
            affiliation /= cp.einsum("fkn->fn", affiliation)[..., None, :]
            quadratic_form = cp.ones(affiliation_shape, dtype=y.real.dtype)
        elif isinstance(initialization, cp.ndarray):
            num_classes = initialization.shape[-2]
            assert num_classes > 1, num_classes
            affiliation_shape = (*independent, num_classes, num_observations)

            # Force same number of dims (Prevent wrong input)
            assert initialization.ndim == len(affiliation_shape), (
                initialization.shape,
                affiliation_shape,
            )

            # Allow singleton dimensions to be broadcasted
            assert initialization.shape[-2:] == affiliation_shape[-2:], (
                initialization.shape,
                affiliation_shape,
            )

            affiliation = cp.broadcast_to(initialization, affiliation_shape)
            quadratic_form = cp.ones(affiliation_shape, dtype=y.real.dtype)
        elif isinstance(initialization, CACGMM):
            # weight[-2] may be 1, when weight is fixed to 1/K
            # num_classes = initialization.weight.shape[-2]
            num_classes = initialization.cacg.covariance_eigenvectors.shape[-3]

            model = initialization
        else:
            raise TypeError("No sufficient initialization.")

        if source_activity_mask is not None:
            assert (
                source_activity_mask.dtype == cp.bool
            ), source_activity_mask.dtype  # noqa
            assert source_activity_mask.shape[-2:] == (num_classes, num_observations), (
                source_activity_mask.shape,
                independent,
                num_classes,
                num_observations,
            )  # noqa

            if isinstance(initialization, cp.ndarray):
                assert source_activity_mask.shape == initialization.shape, (
                    source_activity_mask.shape,
                    initialization.shape,
                )  # noqa

        assert num_classes < 20, f"num_classes: {num_classes}, sure?"
        assert D < 35, f"Channels: {D}, sure?"

        for iteration in range(iterations):
            if model is not None:
                affiliation, quadratic_form, _ = model._predict(
                    y,
                    source_activity_mask=source_activity_mask,
                    affiliation_eps=affiliation_eps,
                )

            model = self._m_step(
                y,
                quadratic_form,
                affiliation=affiliation,
                saliency=saliency,
                hermitize=hermitize,
                covariance_norm=covariance_norm,
                eigenvalue_floor=eigenvalue_floor,
                weight_constant_axis=weight_constant_axis,
            )

        return model

    def fit_predict(
        self,
        y,
        initialization=None,
        num_classes=None,
        iterations=100,
        *,
        saliency=None,
        source_activity_mask=None,
        weight_constant_axis=(-1,),
        hermitize=True,
        covariance_norm="eigenvalue",
        affiliation_eps=1e-10,
        eigenvalue_floor=1e-10,
    ):
        """Fit a model. Then just return the posterior affiliations."""
        model = self.fit(
            y=y,
            initialization=initialization,
            num_classes=num_classes,
            iterations=iterations,
            saliency=saliency,
            source_activity_mask=source_activity_mask,
            weight_constant_axis=weight_constant_axis,
            hermitize=hermitize,
            covariance_norm=covariance_norm,
            affiliation_eps=affiliation_eps,
            eigenvalue_floor=eigenvalue_floor,
        )
        return model.predict(y)

    def _m_step(
        self,
        x,
        quadratic_form,
        affiliation,
        saliency,
        hermitize,
        covariance_norm,
        eigenvalue_floor,
        weight_constant_axis,
    ):
        weight = estimate_mixture_weight(
            affiliation=affiliation,
            saliency=saliency,
            weight_constant_axis=weight_constant_axis,
        )

        if saliency is None:
            masked_affiliation = affiliation
        else:
            masked_affiliation = affiliation * saliency[..., None, :]

        cacg = ComplexAngularCentralGaussianTrainer()._fit(
            y=x[..., None, :, :],
            saliency=masked_affiliation,
            quadratic_form=quadratic_form,
            hermitize=hermitize,
            covariance_norm=covariance_norm,
            eigenvalue_floor=eigenvalue_floor,
        )
        return CACGMM(weight=weight, cacg=cacg)



from setuptools import setup, find_packages

# To use a consistent encoding
from codecs import open
from os import path

here = path.abspath(path.dirname(__file__))

# Get the long description from the README file
with open(path.join(here, "README.md"), encoding="utf-8") as f:
    long_description = f.read()

setup(
    name="cacgmm-gpu",
    version="0.1.0",
    description="GPU-based CACGMM trainer in CuPy",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/desh2608/cacgmm",
    author="Desh Raj",
    author_email="r.desh26@gmail.com",
    keywords="speech enhancement gss",
    packages=find_packages(exclude=["contrib", "docs", "tests"]),  # Required
)



# The functions here are modified from:
# https://github.com/fgnt/pb_bss/blob/master/pb_bss/distribution/cacgmm.py

from dataclasses import dataclass, field

import cupy as cp
from cacgmm.cacg import ComplexAngularCentralGaussian
from cacgmm.utils import log_pdf_to_affiliation, normalize_observation


def logsumexp(a, axis=None):
    a_max = cp.amax(a, axis=axis, keepdims=True)

    if a_max.ndim > 0:
        a_max[~cp.isfinite(a_max)] = 0
    elif not cp.isfinite(a_max):
        a_max = 0

    tmp = cp.exp(a - a_max)

    # suppress warnings about log of zero
    with cp.errstate(divide="ignore"):
        s = cp.sum(tmp, axis=axis, keepdims=False)
        out = cp.log(s)

    a_max = cp.squeeze(a_max, axis=axis)
    out += a_max

    return out


@dataclass
class CACGMM:
    weight: cp.array = None  # (..., K, 1) for weight_constant_axis==(-1,)  (..., 1, K, T) for weight_constant_axis==(-3,)
    cacg: ComplexAngularCentralGaussian = field(
        default_factory=ComplexAngularCentralGaussian
    )

    def predict(self, y, return_quadratic_form=False, source_activity_mask=None):
        assert cp.iscomplexobj(y), y.dtype
        y = normalize_observation(y)  # swap D and T dim
        affiliation, quadratic_form, _ = self._predict(
            y, source_activity_mask=source_activity_mask
        )
        if return_quadratic_form:
            return affiliation, quadratic_form
        else:
            return affiliation

    def _predict(self, y, source_activity_mask=None, affiliation_eps=0.0):
        """
        Note: y shape is (..., D, T) and not (..., T, D) like in predict
        Args:
            y: Normalized observations with shape (..., D, T).
        Returns: Affiliations with shape (..., K, T) and quadratic format
            with the same shape.
        """
        *independent, _, num_observations = y.shape

        log_pdf, quadratic_form = self.cacg._log_pdf(y[..., None, :, :])

        affiliation = log_pdf_to_affiliation(
            self.weight,
            log_pdf,
            source_activity_mask=source_activity_mask,
            affiliation_eps=affiliation_eps,
        )

        return affiliation, quadratic_form, log_pdf

    def log_likelihood(self, y):
        assert cp.iscomplexobj(y), y.dtype
        y = normalize_observation(y)  # swap D and T dim
        affiliation, quadratic_form, log_pdf = self._predict(y)
        return self._log_likelihood(y, log_pdf)

    def _log_likelihood(self, y, log_pdf):
        """
        Note: y shape is (..., D, T) and not (..., T, D) like in log_likelihood
        Args:
            y: Normalized observations with shape (..., D, T).
            log_pdf: shape (..., K, T)
        Returns:
            log_likelihood, scalar
        """
        *independent, channels, num_observations = y.shape

        # log_pdf.shape: *independent, speakers, num_observations

        # first: sum above the speakers
        # second: sum above time frequency in log domain
        log_likelihood = cp.sum(logsumexp(log_pdf, axis=-2))
        return log_likelihood
